import tempfile
from typing import IO, List, Dict, Optional, Union

import pandas

from .api.v3 import (
    create_data_source_from_files,
)
from .data_source import DataSource
from .enums import DataSourceLabel, FileType
from .exceptions import InvalidID
from .firebase_api import cloud_firestore
from .firebase_api.exceptions import NotFound
from .model import Model
from .user import User


class Project:
    """
    ForecastFlow project class
    """

    def __init__(self, user: 'User', project_id: str):
        """
        Instantiate Project object with given project ID

        Args:
            user:
                Owner of project.

            project_id:
                Project ID you want to open.
        """
        self.user = user
        self.project_id = project_id
        self._data_sources: Dict[str, 'DataSource'] = {}
        self._models: Dict[str, 'Model'] = {}

        try:
            document = self._document
        except NotFound:
            raise InvalidID('Given Project ID is not found')

        self.name = document['name']

    @property
    def _document(self) -> dict:
        doc = cloud_firestore.get(
            f"users/{self.user.user_id}/projects/{self.project_id}", self.user.id_token
        )
        return doc['fields']

    def create_data_source(
        self,
        data: Union['pandas.DataFrame', List[Union[str, IO]]],
        name: str,
        label: 'DataSourceLabel',
        description: Optional[str] = None,
        filetype: Optional['FileType'] = None,
        skip_profile: bool = False,
    ) -> 'DataSource':
        """
        Upload file to ForecastFlow and create data source.

        Args:
            data:
                pandas.DataFrame or list of files to upload.

            name:
                Name of data source.

            label:
                Label of data.

            description:
                Description of data source.

            filetype:
                Data file format

        Returns:
            New data source created.
        """
        if isinstance(data, pandas.DataFrame):
            if filetype is None:
                filetype = FileType.CSV
            return self._create_data_source_from_data_frame(
                df=data,
                filetype=filetype,
                name=name,
                label=label,
                description=description,
                skip_profile=skip_profile,
            )
        else:
            if filetype is None:
                raise ValueError('filetype should be specified')
            else:
                return self._create_data_source_from_files(
                    files=data,
                    filetype=filetype,
                    name=name,
                    label=label,
                    description=description,
                    skip_profile=skip_profile,
                )

    def _create_data_source_from_data_frame(
        self,
        df: 'pandas.DataFrame',
        filetype: 'FileType',
        name: str,
        label: 'DataSourceLabel',
        description: Optional[str] = None,
        skip_profile: bool = False,
    ) -> 'DataSource':
        if filetype != FileType.CSV:
            raise ValueError(
                f'{filetype} is not supported for uploading `pandas.DataFrame`.'
            )
        with tempfile.TemporaryFile(mode='rt+') as f:
            df.to_csv(f)
            f.seek(0)
            return self._create_data_source_from_files(
                files=[f],
                filetype=filetype,
                name=name,
                label=label,
                description=description,
                skip_profile=skip_profile,
            )

    def _create_data_source_from_files(
        self,
        files: List[Union[str, IO]],
        filetype: 'FileType',
        name: str,
        label: 'DataSourceLabel',
        description: Optional[str] = None,
        skip_profile: bool = False,
    ) -> 'DataSource':
        did = create_data_source_from_files(
            id_token=self.user.id_token,
            uid=self.user.user_id,
            pid=self.project_id,
            files=files,
            filetype=filetype,
            name=name,
            label=label,
            description=description,
        )
        self._data_sources[did] = DataSource(self, did)
        if not skip_profile:
            self._data_sources[did].profile()
        return self._data_sources[did]

    def get_data_source(self, data_source_id) -> 'DataSource':
        """
        Get data source with given data source id.

        Args:
            data_source_id:
                ID of data source you want to open.

        Returns:
            ForecastFlow data source object with given data source ID.
        """
        if data_source_id not in self._data_sources:
            d = DataSource(self, data_source_id)
            self._data_sources[data_source_id] = d
        return self._data_sources[data_source_id]

    def get_model(self, model_id) -> 'Model':
        """
        Get model with given model ID.

        Args:
            model_id:
                ID of model you want to open.

        Returns:
            ForecastFlow model object with given model ID.
        """
        if model_id not in self._models:
            self._models[model_id] = Model(self, model_id)
        return self._models[model_id]

    @property
    def pid(self) -> str:
        return self.project_id
